from setuptools import find_packages, setup

setup(name='UnCodeGL',
      version='0.0.3',
      description='Library for Coding Theory',
      author= 'Gabriel Octavio Lozano Pinzon',
      author_email="golozanop@unal.edu.co",
      url="https://github.com/Matematikoi/UnCodeGL",
      install_requires=[],
      packages=find_packages(),
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
      ],
      python_requires='>=3.6'
      )